# Drop moves item(s) to the bottom of a queue; Pop moves them to the top! These
# use default method (name of the command) and default method call (just pass in
# the criteria) so there's little code.

from ..command.command import QueueCommand


class DropCommand(QueueCommand):

    name = 'drop'
    key = 'r'
    prompt = "d(r)op"

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        self._add_criteria_arg(parser)

    def call_method(self):
        dropped = self._queue.drop(*self._criteria)
        if len(dropped) == 1:
            self.status = f"Dropped: {dropped[0].description}"
        elif dropped:
            self.status = f"Dropped {str(len(dropped))} Items"
        else:
            self.status = "Dropped nothing"


class PopCommand(QueueCommand):

    name = 'pop'
    key = 'o'
    prompt = "p(o)p"

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        self._add_criteria_arg(parser)

    def call_method(self):
        popped = self._queue.pop(*self._criteria)
        if len(popped) == 1:
            self.status = f"Popped: {popped[0].description}"
        elif popped:
            self.status = f"Popped {str(len(popped))} Items"
        else:
            self.status = "Popped nothing"



from unittest import TestCase
from unittest.mock import Mock
from io import StringIO
from datetime import date as Date

from busy.command.resource_command import ResourceCommand
from busy.queue.todo_queue import TodoQueue

class TestCommandResource(TestCase):

    def test_resource(self):
        q = TodoQueue(items=['a at g', 'b at h'])
        c = ResourceCommand()
        c._queue_ = q
        rx = c.call_method()
        self.assertEqual(rx, 'g')

    # def test_get_without_resource(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a at g\n')
    #         c = Handler(FilesystemRoot(t))
    #         o = c.handle('get-without-resource')
    #         self.assertEqual(o, 'a')



from unittest import TestCase
from unittest.mock import Mock
from unittest.mock import patch
from datetime import date as Date

from busy.command.finish_command import FinishCommand

class TestCommandFinish(TestCase):

    def test_call(self):
        c = FinishCommand()
        c._queue_ = Mock()
        c._queue_.finish.return_value = [[],[],[]]
        c._queue_.list.return_value = [(1,'x')]
        c._namespace = Mock()
        # c._namespace.criteria = []
        c._namespace.yes = True
        c._namespace.timing = "today"
        c._indices_ = [0]
        c.call_method()
        c._queue.finish.assert_called_once()

    def test_confirm_description(self):
        c = FinishCommand()
        c._queue_ = Mock()
        c._queue_.list.return_value = [(1,'x'),(3,'z')]
        c._namespace = Mock()
        c._namespace.yes = None
        c._namespace.queue = None
        c._namespace.criteria = [1,3]
        c._namespace.timing = None
        c._ui = Mock()
        c._ui.get_confirmation.return_value = True
        c._strings_ = []
        with patch('busy.util.date_util.today', lambda : Date(2019,2,11)):
            c.clean_args()
            c._ui.get_confirmation.assert_called_with('Finished on 2019-02-11')

    def test_status_one(self):
        c = FinishCommand()
        c._queue_ = Mock()
        t = Mock()
        t.description = 'foo'
        c._queue_.finish.return_value = [[t],[],[]]
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.yes = True
        c._namespace.timing = "today"
        c._indices_ = [0]
        c.call_method()
        self.assertEqual(c.status, "Finished: foo")

    def test_status_multiple(self):
        c = FinishCommand()
        c._queue_ = Mock()
        c._queue_.finish.return_value = [['',''],[''],['']]
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.yes = True
        c._namespace.timing = "today"
        c._indices_ = [0]
        c.call_method()
        self.assertEqual(c.status, "Finished 2 Items // Added 1 // Repeated 1")

    def test_status_unconfirmed(self):
        c = FinishCommand()
        c._queue_ = Mock()
        c._queue_.finish.return_value = [['',''],[''],['']]
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.yes = None
        c._indices_ = [0]
        c.call_method()
        self.assertEqual(c.status, "Finish operation unconfirmed")

    def test_status_none(self):
        c = FinishCommand()
        c._queue_ = Mock()
        c._queue_.finish.return_value = [[],[],[]]
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.yes = True
        c._indices_ = []
        c.call_method()
        self.assertEqual(c.status, "Finished nothing")


    def test_default_criteria_in_clean_args(self):
        u = Mock()
        u.get_string = lambda d: 'f'
        c = FinishCommand(None, u)
        c._queue_ = Mock()
        c._queue_.list.return_value = []
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.queue = None
        c._namespace.timing = "today"
        c._strings_ = []
        c.clean_args()
        self.assertEqual(c._namespace.criteria, [1])

    def test_default_today(self):
        u = Mock()
        # u.get_string = lambda d: 'f'
        c = FinishCommand(None, u)
        c._queue_ = Mock()
        c._queue_.list.return_value = []
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.queue = None
        c._namespace.timing = None
        c._strings_ = []
        with patch('busy.util.date_util.today', lambda : Date(2019,2,11)):
            c.clean_args()
            self.assertEqual(c._date, Date(2019, 2, 11))




# A Command class handles argparse stuff (ArgumentParser and Namespace) on
# behalf of a method somewhere in the model (i.e. on Queue) and returns text
# that can be used in a UI. A Command class rarely performs an actual operation.
#
# Command classes must have the following class-level properties:
#
# ui - Name of the UI associated with this Command. Default is 'shell', set in
# base class. Inside the command, self._ui refers to the UI object itself.
#
# name - Name of the command. Used in the original Shell UI and also as the name
# of the methods on the queue.
#
# Commands must implement the following methods:
#
# @classmethod set_parser(parser) - Receives an ArgumentParser (actually a
# subparser, but that's irrelevant here), adds arguments to it as appropriate
# for that command, and keeps a reference to it in self.parser.
#
# __init__(root, ui) - Receives the Root and UI; sets self._* for both.
# Implemented in the base class.
#
# execute(namespace) - Perform the command. It might use the namespace from the
# Handler, or it might accept a new Namespace, in the event of interactive UIs.
# The execute method is expected to set a property "status" with a one-line
# string description of what happened.
#
# A Command might implement the following methods:
#
# clean_args() - Examine the self._namespace and use the UI to request any
# args that are omitted. Tends to be command-specific.
#
# They might have the following class attributes:
#
# key - For the Curses UI. What key the user should press.
#
# TODO: Move the base class to __init__.py
# TODO: Make it an ABC

from busy.util.class_family import ClassFamily

class Command(ClassFamily):

    ui = 'shell'

    @classmethod
    def set_parser(self, parser):
        self.parser = parser

    # Convenience method to add an argument to flag for confirmation in the
    # command line itself.

    @classmethod
    def _add_confirmation_arg(self, parser):
        parser.add_argument('--yes', action='store_true')

    def __init__(self, root=None, ui=None):
        self._root = root
        self._ui = ui
        self.status = None

    # Hang onto the namespace that's passed in, for use by other execute
    # methods. Note this might be different from the Handler's Namespace. This
    # one is just the abstract; others should inherit, use super, and actually
    # do something.
    #
    # There is an uncomfortable pattern here. A property is being set in a
    # non-initializer method, which methods later in the lifecycle rely on.

    def execute(self, namespace):
        self._namespace = namespace
        self.clean_args()


    # The clean_args method serves 3 purposes.
    # 
    # (1) If any arguments are missing from the command, ask the user to provide
    # them through the UI. For example, when adding a new item to a queue, if a
    # description is not yet provided, ask the user to provide one.
    # 
    # (2) This is where defaulting happens. Defaults should generally not be
    # applied in the model, but rather here in the command layer, closer to the
    # user.
    #
    # (3) Call for confirmation from the user (using is_confirmed) if
    # appropriate.
    #
    # The default is to do nothing, as this is essentially an abstract method
    # ripe for override.

    def clean_args(self):
        pass


    # Convenience method to find out whether an argument was omitted when the
    # Command was created

    def is_omitted(self, argument):
        if hasattr(self._namespace, argument):
            return not getattr(self._namespace, argument)
        return True


# Abstract base command for commands that work with one queue. A QueueCommand
# knows the reference to the Queue, and takes care of loading and saving with
# the root. A QueueCommand subclass might implement:
#
# call_method() - Call the actual method on the queue. This allows for
# specialized arguments passed into the method. If omitted, the default is to
# pass the criteria from the Namespace.

class QueueCommand(Command):

    default_queue = 'tasks'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('--queue', nargs=1, dest="queue")

    # Convenience method used by some queue commands to add criteria as an
    # argument, to allow for the selection of items from the queue. Some queue
    # commands operate on the whole queue and don't require criteria.

    @classmethod
    def _add_criteria_arg(self, parser):
        parser.add_argument('criteria', action='store', nargs="*")

    # Set the queue if it isn't designated.

    def clean_args(self):
        super().clean_args()
        if self.is_omitted('queue'):
            self._namespace.queue = [self.default_queue]


    # Get the Queue object being acted on. Remember it the first time so it's
    # easy to get the next time.

    @property
    def _queue(self):
        if not hasattr(self, '_queue_'):
            queue_name = self._namespace.queue[0]
            self._queue_ = self._root.get_queue(queue_name)
        return self._queue_


    # Many commands take criteria to get items from a queue. Here are shortcuts
    # to look up the indices and the strings only once.

    @property
    def _indices(self):
        if not hasattr(self, '_indices_'):
            self._indices_ = self._queue.indices(*self._criteria)
        return self._indices_

    @property
    def _strings(self):
        if not hasattr(self, '_strings_'):
            self._strings_ = self._queue.strings(*self._indices)
        return self._strings_


    # Convenience method to figure out whether the command is confirmed, either
    # in the original command or from the UI. Assumes indices have been set
    # based on the criteria.

    def clean_confirmation(self, verb):
        if self.is_omitted('yes'):
            if self._indices:
                value = self._ui.get_confirmation(verb)
            else:
                value = None
        else:
            value = True
        self._namespace.yes = value


    # Queues contain methods with the same name as the commands, so we really
    # just need to get it and call it.

    def execute(self, namespace):
        super().execute(namespace)
        result = self.call_method()
        self._root.save()
        return result

    # The default, which works for many queue commands, is to pass the criteria
    # into the method. This never sets the status, so not useful?

    def call_method(self):
        method = getattr(self._queue, self.name)
        method(*self._namespace.criteria)


    # Convenience method to get the criteria from the namespace, including
    # handling the case when there is none provided

    @property
    def _criteria(self):
        if hasattr(self._namespace, 'criteria'):
            return self._namespace.criteria
        else:
            return []

# Commands that are specific to the task queue inherit here. Ideally we would
# verify that the queue being used is a TodoQueue, but we're going to rely on
# defaults and argparse for that now.

class TodoCommand(QueueCommand):


    # Make sure we're using the tasks queue. Think about whether the "default"
    # queue is the same as the "required" queue for todo operations.

    def clean_args(self):
        super().clean_args()
        if self._namespace.queue != [self.default_queue]:
            message = "Todo commands only operate on the "
            message += f"{self.default_queue} queue."
            raise RuntimeError(message)


from unittest import TestCase
from unittest.mock import Mock

from busy.command.list_command import ListCommand

class TestCommandList(TestCase):

    def test_list_calls(self):
        c = ListCommand()
        c._queue_ = Mock()
        c._queue.list.return_value = []
        c._queue.itemclass.listfmt = ""
        c._namespace = Mock()
        c._namespace.criteria = []
        c.call_method()
        c._queue.list.assert_called_once()

    def test_list_status_1(self):
        c = ListCommand()
        c._queue_ = Mock()
        i = Mock()
        c._queue.list.return_value = [(1,i)]
        c._queue.itemclass.listfmt = ""
        c._namespace = Mock()
        c._namespace.criteria = []
        c.call_method()
        self.assertEqual(c.status, "Listed 1 Item")

    def test_list_status_2(self):
        c = ListCommand()
        c._queue_ = Mock()
        c._queue.itemclass.listfmt = ""
        i = Mock()
        j = Mock()
        c._queue.list.return_value = [(1,i), (2,j)]
        c._namespace = Mock()
        c._namespace.criteria = []
        c.call_method()
        self.assertEqual(c.status, "Listed 2 Items")

    def test_list_status_0(self):
        c = ListCommand()
        c._queue_ = Mock()
        c._queue.itemclass.listfmt = ""
        c._queue.list.return_value = []
        c._namespace = Mock()
        c._namespace.criteria = []
        c.call_method()
        self.assertEqual(c.status, "Listed nothing")

    def test_list_result(self):
        c = ListCommand()
        c._queue_ = Mock()
        i = Mock()
        i.description = "i"
        j = Mock()
        j.description = "j"
        c._queue.list.return_value = [(1,i), (2,j)]
        c._queue.itemclass.listfmt = "{1.description}"
        c._namespace = Mock()
        c._namespace.criteria = []
        r = c.call_method()
        self.assertEqual(r, "     1  i\n     2  j")



from unittest import TestCase
from unittest.mock import Mock

from busy.command.command import QueueCommand

class TestCommand(TestCase):

    def test_queue_command_execute(self):
        c = QueueCommand()
        c._root = Mock()
        c.call_method = lambda: 'a'
        n = Mock()
        n.queue = ['q']
        r = c.execute(n)
        self.assertEqual(r, 'a')


from unittest import TestCase
from unittest.mock import Mock

from busy.command.defer_command import DeferCommand

class TestCommandDefer(TestCase):

    def test_defer(self):
        c = DeferCommand()
        c._queue_ = Mock()
        i = Mock()        
        c._queue.defer.return_value = [i]
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.deferral = "tomorrow"
        c._indices_ = [1]
        c.call_method()
        c._queue.defer.assert_called_once()

    def test_defer_respond_nothing(self):
        c = DeferCommand()
        c._queue_ = Mock()
        c._queue.defer.return_value = []
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.deferral = "tomorrow"
        c._indices_ = []
        c.call_method()
        self.assertEqual(c.status, "Deferred nothing")

    def test_clean_args(self):
        c = DeferCommand(None, Mock())
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.deferral = "tomorrow"
        c._namespace.queue = None
        c._strings_ = []
        c.clean_args()

    def test_input_timing(self):
        u = Mock()
        u.get_string = lambda p, d: 'tomorrow'
        c = DeferCommand(None, u)
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.queue = None
        c._namespace.deferral = None
        c._strings_ = []
        c.clean_args()
        self.assertEqual(c._namespace.deferral, 'tomorrow')

    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\nc\nd')
    #         c = Handler(FilesystemRoot(t))
    #         c.handle('defer','2','--to','2019-09-06')
    #         o = p.read_text()
    #         self.assertEqual(o, 'a\nc\nd\n')
    #         o2 = Path(t, 'plans.txt').read_text()
    #         self.assertEqual(o2, '2019-09-06|b\n')

    # def test_defer_for(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\nc\nd')
    #         c = Handler(FilesystemRoot(t))
    #         c.handle('defer','2','--for','2019-09-06')
    #         o = p.read_text()
    #         self.assertEqual(o, 'a\nc\nd\n')
    #         o2 = Path(t, 'plans.txt').read_text()
    #         self.assertEqual(o2, '2019-09-06|b\n')

    # def test_defer_days(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\nc\nd')
    #         with mock.patch('busy.date_util.today', lambda : Date(2019,2,11)):
    #             c = Handler(FilesystemRoot(t))
    #             c.handle('defer','2','--for','1 day')
    #             o = p.read_text()
    #             self.assertEqual(o, 'a\nc\nd\n')
    #             o2 = Path(t, 'plans.txt').read_text()
    #             self.assertEqual(o2, '2019-02-12|b\n')

    # def test_defer_d(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\nc\nd')
    #         with mock.patch('busy.date_util.today', lambda : Date(2019,2,11)):
    #             c = Handler(FilesystemRoot(t))
    #             c.handle('defer','2','--for','5d')
    #             o = p.read_text()
    #             self.assertEqual(o, 'a\nc\nd\n')
    #             o2 = Path(t, 'plans.txt').read_text()
    #             self.assertEqual(o2, '2019-02-16|b\n')

    # def test_defer_with_input(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\n')
    #         c = Handler(FilesystemRoot(t))
    #         with mock.patch('sys.stdin', StringIO('2019-08-24')):
    #             c.handle('defer')
    #             o = p.read_text()
    #             self.assertEqual(o, 'b\n')
    #             o2 = Path(t, 'plans.txt').read_text()
    #             self.assertEqual(o2, '2019-08-24|a\n')

    # def test_default_tomorrow(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\n')
    #         c = Handler(FilesystemRoot(t))
    #         with mock.patch('busy.date_util.today', lambda : Date(2019,2,11)):
    #             with mock.patch('sys.stdin', StringIO(' ')):
    #                 c.handle('defer')
    #                 o2 = Path(t, 'plans.txt').read_text()
    #                 self.assertEqual(o2, '2019-02-12|a\n')

    # def test_tuesday(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\n')
    #         c = Handler(FilesystemRoot(t))
    #         with mock.patch('busy.date_util.today', lambda : Date(2019,2,14)):
    #             c.handle('defer','--to','tue')
    #             o2 = Path(t, 'plans.txt').read_text()
    #             self.assertEqual(o2, '2019-02-19|a\n')

    # def test_thursday(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\n')
    #         c = Handler(FilesystemRoot(t))
    #         with mock.patch('busy.date_util.today', lambda : Date(2019,2,14)):
    #             c.handle('defer','--to','thursday')
    #             o2 = Path(t, 'plans.txt').read_text()
    #             self.assertEqual(o2, '2019-02-21|a\n')

    # def test_capital_days(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\n')
    #         c = Handler(FilesystemRoot(t))
    #         with mock.patch('busy.date_util.today', lambda : Date(2019,2,14)):
    #             c.handle('defer','--to','THURS')
    #             o2 = Path(t, 'plans.txt').read_text()
    #             self.assertEqual(o2, '2019-02-21|a\n')

    # def test_slashes(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a')
    #         c = Handler(FilesystemRoot(t))
    #         c.handle('defer','--to','2019/09/06')
    #         o2 = Path(t, 'plans.txt').read_text()
    #         self.assertEqual(o2, '2019-09-06|a\n')



from unittest import TestCase
from unittest.mock import Mock
from unittest.mock import patch

from busy.command.add_command import AddCommand


class TestCommandAdd(TestCase):

    def test_call_method(self):
        c = AddCommand()
        c._queue_ = Mock()
        c._namespace = Mock()
        c._namespace.description = 'd'
        c.call_method()
        c._queue.add.assert_called_with('d')

    def test_add_omitted_description(self):
        u = Mock()
        u.get_string = lambda d: 'f'
        c = AddCommand(None, u)
        c._namespace = Mock()
        c._namespace.description = None
        c.clean_args()
        self.assertEqual(c._namespace.description, 'f')


from ..command.command import TodoCommand
from busy.util import date_util


# The Activate command is unusal because any criteria apply to the plans queue,
# not the main tasks queue.


def is_today_or_earlier(plan):
    return plan.date <= date_util.today()

class ActivateCommand(TodoCommand):

    name = 'activate'
    key = 'c'


    # Change in this version: "today" is the default if no criteria are given,
    # not a specific option.

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        self._add_criteria_arg(parser)
        self._add_confirmation_arg(parser)


    def clean_args(self):
        super().clean_args()
        self._ui.output('\n'.join(self._strings))
        self.clean_confirmation("Activate")


    # Indices behave differently, because they apply to the plans queue. We're
    # also going to do the defaulting here.
    #
    # TODO: Create a simpler way to handle the confirmation on another queue.

    @property
    def _indices(self):
        if not hasattr(self, '_indices_'):
            if self.is_omitted('criteria'):
                indices = self._queue.plans.indices(is_today_or_earlier)
                self._indices_ = indices
            else:
                self._indices_ = self._queue.plans.select(*self._criteria)
        return self._indices_

    @property
    def _strings(self):
        if not hasattr(self, '_strings_'):
            self._strings_ = self._queue.plans.strings(*self._indices)
        return self._strings_


    def call_method(self):
        if not self._indices:
            self.status = "Activated nothing"
        elif not self._namespace.yes:
            self.status = "Activate operation unconfirmed"
        else:
            activated = self._queue.activate(*self._indices)
            if len(activated) == 1:
                self.status = f"Activated: {activated[0].description}"
            else:
                self.status = f"Activated {str(len(activated))} Items"



from unittest import TestCase
from unittest.mock import Mock

from busy.command.top_command import TopCommand
from busy.queue.queue import Queue

class TestCommandGet(TestCase):

    def test_call_method(self):
        c = TopCommand()
        c._queue_ = Mock()
        c._namespace = Mock()
        c.call_method()
        c._queue.top.assert_called_once_with()
        self.assertIsNone(c.status)

    def test_call_method_empty_queue(self):
        c = TopCommand()
        c._queue_ = Mock()
        c._queue.top = lambda: None
        c._namespace = Mock()
        c.call_method()
        self.assertTrue(len(c.status) > 10)
    
    def test_deep(self):
        q = Queue(items=['a #b at c --> repeat tomorrow', 'd #e at f'])
        c = TopCommand()
        c._queue_ = q
        x = c.call_method()
        self.assertEqual(x, 'a #b at c --> repeat tomorrow')


from unittest import TestCase
from pathlib import Path
from tempfile import TemporaryDirectory

from busy.queue.todo_queue import TodoQueue
from busy.queue.todo_queue import PlanQueue
from busy.queue.queue import Queue
from busy.model.task import Task
from busy.root.file_system_root import File
from busy.model.item import Item

class TestFile(TestCase):

    def test_load_items(self):
        with TemporaryDirectory() as d:
            p = Path(d) / 'x.txt'
            p.write_text('a\nb\n')
            f = File(p)
            q = f.read()
            self.assertIsInstance(q[0], Item)
            self.assertEqual(q[0].description, 'a')

    def test_load_if_not_there(self):
        with TemporaryDirectory() as d:
            f = File(Path(d) / 'y.txt')
            q = f.read()
            self.assertIsInstance(q, list)
            self.assertEqual(len(q), 0)

    def test_save_items(self):
        with TemporaryDirectory() as d:
            p = Path(d) / 'z.txt'
            f1 = File(p)
            i = Item('a')
            f1.save(i)
            f2 = File(p)
            self.assertEqual(str(f2.read()[0]), 'a')

    # def test_plan_file_format(self):
    #     with TemporaryDirectory() as d:
    #         p = Path(d) / 'plans.txt'
    #         p.write_text('2018-12-01|a\n2018-12-09|b')
    #         q = File(p, PlanQueue).queue
    #         self.assertEqual(q.get().date.month, 12)



import re

from .item import Item
from busy.util import date_util

MARKER = re.compile(r'\s*\-*\>\s*')
REPEAT = re.compile(r'^\s*repeat(?:\s+[io]n)?\s+(.+)\s*$', re.I)
RESOURCE = re.compile(r'\s+at\s+(\S+)')

class Task(Item):

    def __init__(self, description=None):
        super().__init__(description)

    def as_plan(self, date):
        return Plan(self.description, date)

    def as_done(self, date):
        return DoneTask(self._marker_split[0], date)

    def as_followon(self):
        if len(self._marker_split) > 1:
            description = self._marker_split[1]
            if not REPEAT.match(description):
                return Task(description)

    def as_repeat(self):
        if len(self._marker_split) > 1:
            match = REPEAT.match(self._marker_split[1])
            if match:
                date = date_util.relative_date(match.group(1))
                return Plan(self.description, date)

    @property
    def _marker_split(self):
        return MARKER.split(self.description, maxsplit=1)


    # A resource is meant to be a URL that will be useful for a command. It's
    # designated in the task syntax by "at" followed by a word. The idea is that
    # the resource could be lifted and browed easily to start on a task.

    @property
    def resource(self):
        match = RESOURCE.search(self.description)
        if match:
            return match.group(1)
        else:
            return ""


    # TODO: There must be a more elegant way to parse it

    @property
    def base(self):
        first = self._marker_split[0]
        split = RESOURCE.split(first, maxsplit=1)
        if len(split) > 1:
            without_resource = split[0] + split[2]
        else:
            without_resource = first
        return self._base(without_resource)
    
    @property
    def tags(self):
        first = self._marker_split[0]
        return self._tags(first)



class Plan(Item):

    schema = ['date', 'description']
    listfmt = "{1.date:%Y-%m-%d}  {1.description}"

    def __init__(self, description=None, date=None):
        super().__init__(description)
        self._date = date_util.absolute_date(date)

    @property
    def date(self):
        return self._date

    def as_todo(self):
        return Task(self.description)


class DoneTask(Item):

    schema = ['date', 'description']
    listfmt = "{1.date:%Y-%m-%d}  {1.description}"

    def __init__(self, description=None, date=None):
        super().__init__(description)
        self._date = date_util.absolute_date(date)

    @property
    def date(self):
        return self._date



from unittest import TestCase
from unittest.mock import Mock

from busy.command.activate_command import ActivateCommand
from busy.model.task import Task

class TestCommandActivate(TestCase):

    def test_call(self):
        c = ActivateCommand()
        c._queue_ = Mock()
        c._queue.activate.return_value = []
        c._namespace = Mock()
        c._indices_ = [2]
        c.call_method()
        c._queue.activate.assert_called_with(2)


    def test_clean_args_today(self):
        c = ActivateCommand(None, Mock())
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.queue = None
        c._queue_ = Mock()
        c._queue_.plans.indices = Mock()
        c._queue_.plans.indices.return_value = [1, 3]
        c._strings_ = []
        c.clean_args()
        self.assertEqual(c._indices, [1, 3])




from unittest import TestCase
from unittest.mock import Mock

from busy.command.curses_command import CursesCommand

class TestCommandCurses(TestCase):

    def test_execute(self):
        u = Mock()
        c = CursesCommand(None, u)
        n = Mock()
        c.execute(n)
        u.start.assert_called_once_with()




from ..command.command import QueueCommand


class DeleteCommand(QueueCommand):

    name = 'delete'
    key = 'd'
    prompt = 'd)elete'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('--yes', action='store_true')
        self._add_criteria_arg(parser)


    # TODO: Move more of this into clean_confirmation
    # TODO: Make a convenience method to set the default

    def clean_args(self):
        super().clean_args()
        if self.is_omitted('criteria'):
            self._namespace.criteria = [1]
        self.clean_confirmation("Delete")

    # Assume the indices have been already set, before confirmation.

    def call_method(self):
        if self._namespace.yes:
            deleted = self._queue.delete(*self._indices)
            if len(deleted) == 1:
                item = deleted[0]
                self.status = f"Deleted: {item.description}"
            elif deleted:
                self.status = f"Deleted {len(deleted)} Items"
            else:
                self.status = "Deleted nothing"
        else:
            self.status = "Delete operation unconfirmed"

    def execute_on_queue(self, parsed, queue):
        itemlist = queue.list(*parsed.criteria or [1])
        indices = [i[0]-1 for i in itemlist]
        if self.confirmed(parsed, itemlist, 'Delete'):
            queue.delete_by_indices(*indices)



from unittest import TestCase
from unittest.mock import Mock

from busy.command.delete_command import DeleteCommand

class TestCommandDelete(TestCase):

    # def test_delete(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\nc\nd')
    #         c = Handler(FilesystemRoot(t))
    #         c.setUI(ShellUI(c))
    #         c.handle('delete','--yes','3-')
    #         o = p.read_text()
    #         self.assertEqual(o, 'a\nb\n')

    def test_call_method(self):
        c = DeleteCommand()
        c._queue_ = Mock()
        c._queue.delete.return_value = []
        c._namespace = Mock()
        c._indices_ = []
        c._namespace.yes = True
        c.call_method()
        c._queue.delete.assert_called_once()

    def test_status(self):
        c = DeleteCommand()
        c._queue_ = Mock()
        c._queue.delete.return_value = [[''],['']]
        c._namespace = Mock()
        c._namespace.yes = True
        c._indices_ = [0, 1]
        c.call_method()
        self.assertEqual(c.status, "Deleted 2 Items")

    def test_obey_confirmation(self):
        c = DeleteCommand()
        c._queue_ = Mock()
        c._queue.delete.return_value = [[''],['']]
        c._namespace = Mock()
        c._namespace.yes = False
        c._indices_ = [0, 1]
        c.call_method()
        self.assertEqual(c.status, "Delete operation unconfirmed")

    def test_clean_args_confirms(self):
        u = Mock()
        c = DeleteCommand(None, u)
        c._queue_ = Mock()
        c._queue.list.return_value = [(1,'x')]
        c._namespace = Mock() 
        c._namespace.yes = None
        c._namespace.criteria = []
        c.clean_args()
        u.get_confirmation.assert_called_with("Delete")

    def test_clean_args_sets_yes(self):
        u = Mock()
        u.get_confirmation.return_value = True
        c = DeleteCommand(None, u)
        c._queue_ = Mock()
        c._queue.list.return_value = [(1,'x')]
        c._namespace = Mock()
        c._namespace.criteria = []
        c._namespace.yes = None
        c.clean_args()
        self.assertEqual(c._namespace.yes, True)


    # def test_delete_with_input_confirmation_yes(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\nc\nd')
    #         c = Handler(FilesystemRoot(t))
    #         c.setUI(ShellUI(c))
    #         with mock.patch('sys.stdin', StringIO('Y')):
    #             c.handle('delete','3-')
    #             o = p.read_text()
    #             self.assertEqual(o, 'a\nb\n')

    # def test_delete_with_input_confirmation_no(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\nc\nd')
    #         c = Handler(FilesystemRoot(t))
    #         c.setUI(ShellUI(c))
    #         with mock.patch('sys.stdin', StringIO('no')):
    #             c.handle('delete','3-')
    #             o = p.read_text()
    #             self.assertEqual(o.strip(), 'a\nb\nc\nd')

    # def test_delete_outputs_before_confirmation(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\n')
    #         c = Handler(FilesystemRoot(t))
    #         c.setUI(ShellUI(c))
    #         o = StringIO()
    #         with mock.patch('sys.stdin', StringIO('Y')):
    #             with mock.patch('sys.stdout', o):
    #                 c.handle('delete', '1')
    #                 self.assertEqual(o.getvalue(), 'a\nDelete? (Y/n) ')

    # def test_delete_defaults_to_first_task_only(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\n')
    #         c = Handler(FilesystemRoot(t))
    #         c.setUI(ShellUI(c))
    #         with mock.patch('sys.stdin', StringIO('Y')):
    #             c.handle('delete')
    #             o = p.read_text()
    #             self.assertEqual(o, 'b\n')



from unittest import TestCase
from unittest.mock import patch
from unittest.mock import Mock
from io import StringIO

from busy.ui.shell_ui import ShellUI

class TestShellUI(TestCase):

    def test_get_string(self):
        u = ShellUI()
        with patch('sys.stdin', StringIO('g')):
            v = u.get_string()
            self.assertEqual(v, 'g')

    @patch('sys.stdin', StringIO('Y'))
    def test_get_confirmation(self):
        u = ShellUI()
        o = StringIO()
        with patch('sys.stdout', o):
            v = u.get_confirmation('A')
            self.assertEqual(o.getvalue(), 'A? (Y/n) ')



# General-purpose queue of items. Method calls on queue use indices that start
# at 1.
#
# TODO: Make Queue not a ClassFamily
# TODO: Queue belongs under model

import random
from pathlib import Path

from busy.selector import Selector
from busy.model.item import Item
from busy.util.class_family import ClassFamily


class Queue(ClassFamily):

    itemclass = Item

    def __init__(self, root=None, items=[]):
        self.root = root
        self.changed = False
        self._items = [self.itemclass.create(i) for i in items if i]


    # Take "criteria" (such as tags, ranges of numbers, etc) and return the
    # indices of the resulting items. In other parts of the class and
    # subclasses, operations need to know what items to act on.
    #
    # Both "criteria" and "indices" start with 1, so have to be adjusted
    # internally here when handling the _items list.
    
    def indices(self, *criteria):
        selector = Selector(criteria)
        return [i+1 for i in selector.indices(self._items)]

    def items(self, *indices):
        return [self._items[i-1] for i in indices]


    # Every item has a string representation; this method returns it
    # TODO: I removed the default to all. Is that OK?

    def strings(self, *indices):
        return [str(self._items[i-1]) for i in indices]

    # Return a list of tuples (index, item) based on criteria. Remember that
    # criteria start with an index of 1, but indices start at zero.
    # TODO: Eliminate this one

    def list(self, *criteria):
        return [(i, self._items[i-1]) for i in self.indices(*criteria)]


    def all(self):
        return self._items

    def count(self):
        return len(self._items)


    # Add new items. Always makes them the right class. Also does inserts.

    def add(self, *items, index=None):
        newitems = [self.itemclass.create(i) for i in items if i]
        index = len(self._items) if index is None else index
        self._items[index:index] = newitems
        self.changed = True
        return newitems


    # Replace existing items at the indices provided. Also inserts if the
    # indices run out. Does not create items; expects them to already exist.
    # Indices start at 1, so have to be adjusted going into the _items list.
    # Unlike other methods, it takes raw lists, not unpacked. Does not return
    # anything since the calling code already has the new items. Because the
    # list pop operations are destructive, act on copies of the lists, so the
    # calling code can still use them afterwards.
    
    def replace(self, indices, newvalues):
        _indices = indices.copy()
        _newvalues = newvalues.copy()
        while _newvalues and _indices:
            self._items[_indices.pop(0)-1] = _newvalues.pop(0)
        while _indices:
            del self._items[_indices.pop()-1]
        self._items.extend(_newvalues)
        self.changed = True


    # Used to be called "get" and maybe that was better. Retrieve one item. Used
    # by the "top" command, which is set to always retrieve the first item.

    def top(self, index=1):
        return self._items[index-1] if self._items else None


    def _split(self, *criteria):
        return self._split_by_indices(*self.indices(*criteria))

    def _split_by_indices(self, *indices):
        inlist = [t for i, t in enumerate(self._items) if i+1 in indices]
        outlist = [t for i, t in enumerate(self._items) if i+1 not in indices]
        return (inlist, outlist)


    def pop(self, *criteria):
        if not criteria:
            criteria = [len(self._items)]
        hilist, lolist = self._split(*criteria)
        self._items = hilist + lolist
        self.changed = (len(hilist) > 0)
        return hilist

    def drop(self, *criteria):
        if not criteria:
            criteria = [1]
        lolist, hilist = self._split(*criteria or [1])
        self._items = hilist + lolist
        self.changed = True
        return lolist

    def delete(self, *indices):
        killlist, keeplist = self._split_by_indices(*indices)
        self._items = keeplist
        self.changed = True
        return killlist


    # TODO: Remove the manage operation. Just use list and replace.

    # def manage(self, *criteria, editor):
    #     itemlist = self.list(*criteria)
    #     indices = [i[0]-1 for i in itemlist]
    #     items = [i[1] for i in itemlist]
    #     new_items = editor(self.itemclass, *items)  # move this out to the command and/or UI level
    #     self.replace(indices, new_items)


    # TODO: Bring back the shuffle command

    def shuffle(self, *criteria):
        itemlist = self.list(*criteria)
        indices = [i[0]-1 for i in itemlist]
        items = [i[1] for i in itemlist]
        random.shuffle(items)
        self.replace(indices, items)


    # Get all the tags for the queue. Neat Python trick: return a Set!

    def tags(self):
        master = [i.tags for i in self._items]
        return {item for sublist in master for item in sublist}



import readline

def start():

    # This doesn't work
    readline.parse_and_bind('"\C-s": beginning-of-line')

    # This does work
    readline.insert_text("Hello")

# readline.set_startup_hook(start)
# try:
#     x = input("> ")
# finally:
#     readline.set_startup_hook()
# print(x)


from unittest import TestCase
from unittest.mock import Mock

from busy.command.base_command import BaseCommand
from busy.queue.todo_queue import TodoQueue

class TestBaseCommand(TestCase):

    def test_it(self):
        q = TodoQueue(items=['a #b at c --> d','e'])
        c = BaseCommand()
        c._namespace = Mock()
        c._namespace.criteria = []
        c._queue_ = q
        x = c.call_method()
        self.assertEqual(x, 'a')



from curses import echo, noecho
from busy.util import date_util

from ..command.command import TodoCommand


class DeferCommand(TodoCommand):

    name = 'defer'
    key = 'f'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('--deferral', '--to', '--for')
        self._add_criteria_arg(parser)


    def clean_args(self):
        super().clean_args()
        if self.is_omitted('criteria'):
            self._namespace.criteria = [1]
        self._ui.output('\n'.join(self._strings))
        if self.is_omitted('deferral'):
            deferral = self._ui.get_string("Deferral", "tomorrow")
            self._namespace.deferral = deferral
        self.clean_confirmation(f"Defer to {self._date}")

    @property
    def _date(self):
        if not hasattr(self, '_date_'):
            self._date_ = date_util.relative_date(self._namespace.deferral)
        return self._date_


    def call_method(self):
        if not self._indices:
            self.status = "Deferred nothing"
        elif not self._namespace.yes:
            self.status = "Defer operation unconfirmed"
        else:
            deferred = self._queue.defer(*self._indices, date=self._date)
            if len(deferred) == 1:
                item = deferred[0]
                self.status = f"Deferred: {item.description} // To: {item.date}"
            else:
                item = deferred[0]
                self.status = f"Deferred {str(len(deferred))} Items to {item.date}"



from unittest import TestCase
from unittest.mock import Mock
from unittest.mock import patch

from busy.command.drop_and_pop_command import DropCommand
from busy.command.drop_and_pop_command import PopCommand

class TestCommandDropAndPop(TestCase):

    def test_drop(self):
        c = DropCommand()
        c._queue_ = Mock()
        c._queue.drop.return_value = []
        c._namespace = Mock()
        c._namespace.criteria = []
        c.call_method()
        c._queue.drop.assert_called_once_with()

    def test_drop_respond_nothing(self):
        c = DropCommand()
        c._queue_ = Mock()
        c._queue.drop.return_value = []
        c._namespace = Mock()
        c._namespace.criteria = []
        c.call_method()
        self.assertEqual(c.status, "Dropped nothing")

    def test_drop_respond_count(self):
        c = DropCommand()
        c._queue_ = Mock()
        c._queue.drop.return_value = [Mock(), Mock(), Mock()]
        c._namespace = Mock()
        c._namespace.criteria = []
        c.call_method()
        self.assertEqual(c.status, "Dropped 3 Items")

    def test_drop_respond_description(self):
        c = DropCommand()
        c._queue_ = Mock()
        i = Mock()
        i.description = 'd'
        c._queue.drop.return_value = [i]
        c._namespace = Mock()
        c._namespace.criteria = []
        c.call_method()
        self.assertEqual(c.status, "Dropped: d")


    def test_drop_with_criteria(self):
        c = DropCommand()
        c._queue_ = Mock()
        c._queue.drop.return_value = [Mock(), Mock(), Mock()]
        c._namespace = Mock()
        c._namespace.criteria = [2]
        c.call_method()
        c._queue.drop.assert_called_once_with(2)

    def test_pop(self):
        c = PopCommand()
        c._queue_ = Mock()
        c._queue.pop.return_value = []
        c._namespace = Mock()
        c._namespace.criteria = [2,12]
        c.call_method()
        c._queue.pop.assert_called_once_with(2, 12)





    # def test_output(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\n')
    #         c = Handler(FilesystemRoot(t))
    #         o = c.handle('drop','1')
    #         self.assertIsNone(o)



# Mark a task as finished by moving it to the done queue.

from ..command.command import TodoCommand
from busy.util.date_util import today
from busy.util.date_util import relative_date

class FinishCommand(TodoCommand):

    name = 'finish'
    key = 'n'


    # TODO: Add an option for the date, including e.g. "yesterday"

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('--timing', '--on')
        self._add_confirmation_arg(parser)
        self._add_criteria_arg(parser)


    # TODO: Use relative if a date is specified (for now we are just using today
    # in all cases).

    def clean_args(self):
        super().clean_args()
        if self.is_omitted('criteria'):
            self._namespace.criteria = [1]
        self._ui.output('\n'.join(self._strings))
        if self.is_omitted('timing'):
            self._namespace.timing = "today"
        self.clean_confirmation(f"Finished on {self._date}")


    @property
    def _date(self):
        if not hasattr(self, '_date_'):
            self._date_ = relative_date(self._namespace.timing)
        return self._date_

    # TODO: Move the listing and confirmation into clean_args.
    # TODO: Shouldn't self._criteria reference self._namespace?

    def call_method(self):
        if not self._indices:
            self.status = "Finished nothing"
        elif not self._namespace.yes:
            self.status = "Finish operation unconfirmed"
        else:
            finished, added, deferred = self._queue.finish(*self._indices, date=self._date)
            if len(finished) == 1:
                status = f"Finished: {finished[0].description}"
                if added:
                    status += f" // Added: {added[0].description}"
                if deferred:
                    status += f" // Repeating on {deferred[0].date}"
                self.status = status
            else:
                status = f"Finished {str(len(finished))} Items"
                if added:
                    status += f" // Added {str(len(added))}"
                if deferred:
                    status += f" // Repeated {str(len(deferred))}"
                self.status = status



from unittest import TestCase

from busy.model.item import Item

class TestItem(TestCase):

    def test_tags(self):
        i = Item('a #b')
        tx = i.tags
        self.assertEqual(tx, {'b'})

    def test_tags_no_repeat(self):
        i = Item('a #b #b')
        tx = i.tags
        self.assertEqual(tx, {'b'})

    def test_base(self):
        i = Item('a b #c d e')
        w = i.base
        self.assertEqual(w, 'a b d e')



# The Handler is the core control class for a single execution of the busy
# program. It's called when busy is run, and there are some functions to get
# things configured and started. Then one Handler is instantiated. This is the
# only place that imports the UI, Command, and Root classes. So the handler
# instance is passed around, and other classes can access what they need from
# there. Keeping the imports in one place allows the class families to be more
# modular.
#
# A Handler instance provides the following properties. For the most part, these
# properties are how commands access other class families. Remember that the
# Namespace from the Handler is the original Namespace, so use the one passed in
# to the execute() method of the Command for actual usage in an interactive UI.
# Also, the Handler doesn't hold a queue property, because it is only used by
# certain commands. Instead, get your queue from your root yourself!
#
# namespace - The argparse Namespace from the original parsing of the command
# line
#
# root - The Root object for retrieval and storage of items. This is the main
# way that a command gets its root.
#
# ui - The UI instance being implemented. This is the main way a command gets
# its UI.
#
# command - The original Command instance being implemented. Interactive UIs
# might instantiate other Commands.


from argparse import ArgumentParser
import sys
from pathlib import Path
from sys import argv

from busy.command.command import Command
from busy.ui.ui import UI
from busy.root.file_system_root import FileSystemRoot
from busy.util.python_version import confirm_python_version

PYTHON_VERSION = (3, 6, 5)

def main(*args):
    if not args:
        args = argv[1:]
    confirm_python_version(PYTHON_VERSION) 
    parser = get_parser()      
    namespace = parser.parse_args(args)
    try:
        handler = Handler(namespace)
        handler.execute()
    except (RuntimeError, ValueError) as error:
        print(f"Error: {str(error)}")

def get_parser():
    parser = ArgumentParser(prog='busy')
    parser.add_argument('--root', action='store', \
        help='file system path to busy data', \
        default=None)
    subparsers = parser.add_subparsers(dest='command')
    for command in Command.family_members('name'):
        key = command.get_member_attr('key')
        aliases = [key] if key else []
        subparser = subparsers.add_parser(command.name, aliases=aliases)
        command.set_parser(subparser)
    return parser


class Handler():

    def __init__(self, namespace):
        self.namespace = namespace
        self.root = FileSystemRoot(namespace.root)
        command_class = Command.family_member('name', namespace.command)
        if not command_class:
            raise RuntimeError(f"Unknown command {namespace.command}")
        self.ui = UI.family_member('name', command_class.ui)(self)
        self.command = command_class(self.root, self.ui)

    # Handle the command and output its final status

    def execute(self):
        result = self.command.execute(self.namespace)
        if result:
            print(result)
        if self.command.status:
            print(self.command.status)


    # Some useful methods so other components don't need to import from
    # different class families
    
    def get_command_values(self, *attributes):
        return Command.family_attrs(*attributes)


    # A method for commands within interactive UIs. Use our own root but the UI
    # is passed in. Returns None if attribute value doesn't resolve.
    
    def get_command(self, attribute, value, ui):
        member = Command.family_member(attribute, value)
        if member:
            return member(self.root, ui)




# The UI to execute one command passed in through the shell. There will be
# limited interactivity, if the user omits an argument on the command line, but
# otherwise this is a run and done situation.
#
# The code for argparse lives at https://github.com/python/cpython/blob/main/Lib/argparse.py


from argparse import ArgumentParser
from io import StringIO

from ..ui.ui import TerminalUI


class ShellUI(TerminalUI):

    name = "shell"

    def output(self, intro=""):
        if intro:
            print(intro)

    def get_string(self, prompt='', default=""):
        full_prompt = self.full_prompt(prompt, default)
        value = input(full_prompt) or default
        return value

    def get_confirmation(self, verb):
        confirmed = input(f'{verb}? (Y/n) ').startswith('Y')
        return confirmed



from unittest import TestCase
from unittest.mock import Mock
from unittest.mock import patch

from busy.command.edit_command import EditCommand
from busy.queue.queue import Queue


class TestCommandEdit(TestCase):
    
    def test_manage_one_status(self):
        i = Mock()
        i.description = 'k'
        u = Mock()
        u.edit_items.return_value = [i]
        c = EditCommand(None, u)
        q = Queue(items=['a #b at c --> d','e'])
        c._namespace = Mock()
        c._namespace.criteria = [1]
        c._indices_ = [1]
        c._queue_ = q
        c.call_method()
        self.assertEqual(c.status, 'Edited: k')


    # def test_clean_args_gets_list(self):
    #     c = EditCommand(None, Mock())
    #     c._queue_ = Mock()
    #     c._queue_.list.return_value = [(1,'f')]
    #     c._namespace = Mock()
    #     c._namespace.criteria = None
    #     c.clean_args()
    #     self.assertEqual(c._indices, [0])

    # def test_clean_args_does_no_output(self):
    #     u = Mock()
    #     c = EditCommand(None, u)
    #     c._queue_ = Mock()
    #     c._queue_.list.return_value = [(1,'f')]
    #     c._namespace = Mock()
    #     c._namespace.criteria = None
    #     c.clean_args()
    #     u.output.assert_not_called()

    # def test_edit_text_called(self):
    #     u = Mock()
    #     c = EditCommand(None, u)
    #     c._queue_ = Mock()
    #     c._queue_.list.return_value = [(1,'f')]
    #     c._namespace = Mock()
    #     c._namespace.criteria = []
    #     c.call_method()
    #     u.edit_items.assert_called()


        # with TemporaryDirectory() as t:
        #     p = Path(t, 'tasks.txt')
        #     p.write_text('a\n')
        #     with mock.patch('busy.editor', lambda x: 'b\n'):
        #         c = Handler(FilesystemRoot(t))
        #         o = c.handle('manage')
        #         f = p.read_text()
        #         self.assertEqual(f, 'b\n')

    # def test_manage_includes_newline_at_end(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\n')
    #         m = Mock()
    #         m.return_value = 'b\n'
    #         with mock.patch('busy.editor', m):
    #             c = Handler(FilesystemRoot(t))
    #             o = c.handle('manage')
    #             m.assert_called_with('a\n')

    # def test_leave_tasks_in_place(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\n')
    #         with mock.patch('busy.editor', lambda x: 'a\n'):
    #             c = Handler(FilesystemRoot(t))
    #             o = c.handle('manage', '1')
    #             f = p.read_text()
    #             self.assertEqual(f, 'a\nb\n')

    # def test_last_record(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('a\nb\n')
    #         with mock.patch('busy.editor', lambda x: 'c\n'):
    #             c = Handler(FilesystemRoot(t))
    #             o = c.handle('manage', '-')
    #             f = p.read_text()
    #             self.assertEqual(f, 'a\nc\n')



from unittest import TestCase
from unittest.mock import Mock

from busy.command.queues_command import QueuesCommand

class TestCommandQueues(TestCase):

    def test_queues(self):
        r = Mock()
        r.queue_names = ['a','b']
        c = QueuesCommand(r)
        x = c.execute(None)
        self.assertEqual(x, 'a\nb')



# A generic item from a queue, with a description and possibly tags

from csv import DictReader
from csv import DictWriter


class Item:

    schema = ['description']
    listfmt = "{1.description}"

    def __init__(self, description=None):
        assert isinstance(description, str)
        assert description
        self._description = description

    def __str__(self):
        return self._description

    @property
    def description(self):
        return self._description

    @classmethod
    def create(self, value):
        if isinstance(value, self):
            return value
        elif isinstance(value, str):
            return self(value)
        elif isinstance(value, dict):
            return self(**value)


    # A convenience method to get the tags from a string (description) as a set
    #
    # TODO: Do this with a regexp

    def _tags(self, string):
        words = string.split()
        return {w[1:].lower() for w in words if w.startswith('#')}


    # A convenience method to get the non-tags from a string (description) as a
    # string

    def _base(self, string):
        words = string.split()
        return ' '.join([w for w in words if not w.startswith('#')])


    # A property that is just the tags

    @property
    def tags(self):
        return self._tags(self.description)

    @property
    def base(self):
        return self._base(self.description)

# Read and write a set of items to or from a file-like object. This could be
# part of a root but could also be a tempfile.

def read_items(fileish, itemclass=Item):
    reader = DictReader(fileish, itemclass.schema, delimiter="|")
    return [itemclass.create(i) for i in reader if i]

def write_items(fileish, *items):
    schema = items[0].schema
    writer = DictWriter(fileish, schema, delimiter="|")
    for item in items:
        values = dict([(f, getattr(item, f)) for f in schema])
        writer.writerow(values)




from .command import QueueCommand
import busy


class EditCommand(QueueCommand):

    name = 'edit'
    key = 'e'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        self._add_criteria_arg(parser)

    def clean_args(self):
        super().clean_args()
        if self.is_omitted('criteria'):
            self._namespace.criteria = [1]


    # TODO: Methods to get indices and items as separate lists

    def call_method(self):
        if not self._indices:
            self.status = "Edited nothing"
        else:
            itemclass = self._queue.itemclass
            old = self._queue.items(*self._indices)
            new = self._ui.edit_items(itemclass, *old)
            self._queue.replace(self._indices, new)
            if len(new) == 1:
                status = f"Edited: {new[0].description}"
            else:
                status = f"Edited {str(len(new))} Items"
            self.status = status



from unittest import TestCase
from unittest.mock import Mock
from unittest.mock import patch
# from tempfile import TemporaryDirectory
# from pathlib import Path
# from io import StringIO
# from datetime import date as Date

from busy.command.command import QueueCommand

class TestCommandQueue(TestCase):

    @patch('busy.command.command.QueueCommand.is_omitted', lambda s, a: True)
    def test_omitted_args_default(self):
        c = QueueCommand()
        n = Mock()
        c._namespace = n
        c.clean_args()
        self.assertEqual(['tasks'], c._namespace.queue)

    @patch('busy.command.command.QueueCommand.is_omitted', lambda s, a: False)
    def test_omitted_args_not_default(self):
        c = QueueCommand()
        n = Mock()
        n.queue = ['g']
        c._namespace = n
        c.clean_args()
        self.assertEqual(['g'], c._namespace.queue)

    # These are more like integration tests

    # def test_get_from(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'a.txt')
    #         p.write_text('b\n')
    #         c = Handler(FilesystemRoot(t))
    #         o = c.handle('top','--queue','a')
    #         self.assertEqual(o, 'b')

    # def test_default_queue(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'tasks.txt')
    #         p.write_text('b\n')
    #         c = Handler(FilesystemRoot(t))
    #         o = c.handle('get')
    #         self.assertEqual(o, 'b')

    # def test_add(self):
    #     with TemporaryDirectory() as t:
    #         c = Handler(FilesystemRoot(t))
    #         with mock.patch('sys.stdin', StringIO('g')):
    #             c.handle('add','--queue','j')
    #             x = Path(t, 'j.txt').read_text()
    #             self.assertEqual(x, 'g\n')

    # def test_drop(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'u.txt')
    #         p.write_text('a\nb\nc\nd')
    #         c = Handler(FilesystemRoot(t))
    #         c.handle('drop','2','4','--queue','u')
    #         o = p.read_text()
    #         self.assertEqual(o, 'a\nc\nb\nd\n')

    # def test_delete(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'w.txt')
    #         p.write_text('a\nb\nc\nd')
    #         c = Handler(FilesystemRoot(t))
    #         c.handle('delete','--yes','3-', '--queue', 'w')
    #         o = p.read_text()
    #         self.assertEqual(o, 'a\nb\n')

    # def test_manage(self):
    #     with TemporaryDirectory() as t:
    #         p = Path(t, 'y.txt')
    #         p.write_text('a\n')
    #         with mock.patch('busy.editor', lambda x: 'b\n'):
    #             c = Handler(FilesystemRoot(t))
    #             c.handle('manage', '--queue', 'y')
    #             f = p.read_text()
    #             self.assertEqual(f, 'b\n')



from unittest import TestCase

from busy.ui.ui import UI
from busy.queue.queue import Queue
from busy.handler import Handler
from busy.command.command import Command
from busy.command.finish_command import FinishCommand


class TestClassFamilies(TestCase):

    def test_import_ui(self):
        ui_classes = UI.__subclasses__()
        self.assertEqual(len(ui_classes), 1)

    def test_import_queue(self):
        queue_classes = Queue.family_members()
        self.assertEqual(len(queue_classes), 4)

    def test_import_command(self):
        classes = Command.family_members('name')
        self.assertTrue(FinishCommand in classes)
    
    def test_specific_command(self):
        finish_class = Command.family_member('name', 'finish')
        self.assertEqual(finish_class, FinishCommand)



from unittest import TestCase

from test.data_class_family.atriarch import Atriarch

class TestClassFamily(TestCase):

    def test_family_children(self):
        cx = Atriarch.family_children()
        self.assertEqual(len(cx), 1)

    


from unittest import TestCase
from unittest.mock import Mock
from unittest.mock import patch

from busy.handler import Handler

class TestHandler(TestCase):

    def test_init_passing_root_and_ui_to_class(self):
        r = Mock()
        n = Mock()
        n.root = "r"
        n.command = "c"
        with patch('busy.handler.FileSystemRoot', lambda m: r):
            with patch('busy.handler.Command') as C:
                cc = Mock()
                cc.ui = None
                C.family_member = lambda a, b: cc
                with patch('busy.handler.UI') as U:
                    h = Handler(n)
                    cc.assert_called_with(r, h.ui)

    @patch('busy.handler.FileSystemRoot', Mock())
    def test_get_command(self):
        n = Mock()
        n.command = 'c'
        u = Mock()
        with patch('busy.handler.Command') as C:
            with patch('busy.handler.UI') as U:
                cc = Mock()
                cc.ui = None
                C.family_member = lambda a, b: cc
                # U.family_member = Mock()
                h = Handler(n)
                C.family_member = Mock()
                c = h.get_command('k', 'a', u)
                C.family_member.assert_called_with('k','a')




from ..command.command import TodoCommand


# Get the resource (a URL) for the top entry in a queue

class ResourceCommand(TodoCommand):

    name = "resource"

    def call_method(self):
        return str(self._queue.resource() or '')



# Abstract classes for UIs. The idea here is to allow many different user
# interfaces (shell, curses, Slack, etc) to drive the same data model without
# the data model needing specific UI knowledge.
#
# The UI keeps a reference to the Handler. The Handler will bring a Command with
# it, and this is the main root command. In the case of a Shell-type UI, that is
# the command to be executed. In the case of an interactive UI, it's really just
# a placeholder; the UI will instantiate other commands as it proceeds. But the
# Handler won't know about those.
#
# Commands might call back to the UI for confirmations or arguments that are
# previously omitted, using the get_ methods.

#
# UI end classes must implement the following interface:
#
# __init__(handler): Takes the handler and hangs on to it. Implemented in the
# main base class.
#
# start(): No arguments. Actually performs the operation of the UI. It might be
# short running (in the case of a shell UI) or long-running (in the case of an
# interactive UI).
#
# output(intro=""): Output some multi-line text explaining an action, usually a
# list of items being acted upon.
#
# get_string(prompt, default=""): For arguments that are omitted, get a string
# from the user. The prompt is just a word (like "Description") telling the user
# what to input.
#
# get_confirmation(verb): For delete-oriented commands to
# confirm with the user before proceeding. Verb is what we're asking the user to
# confirm. Description can be multiple lines, as with get_string. Returns a
# boolean saying whether the action is confirmed.

import subprocess
from tempfile import NamedTemporaryFile
from pathlib import Path
import shutil
import os
import re
from io import StringIO

from busy.model.item import write_items
from busy.model.item import read_items
from busy.util.class_family import ClassFamily

class UI(ClassFamily):

    def __init__(self, handler=None):
        self.handler = handler
    
    def start(self):
        pass


# Terminal UIs include Shell and Curses, since both will rely on a
# terminal-based editor for the e.g. Manage command.

class TerminalUI(UI):

    # A convenience method to get a full textual prompt for a string input

    def full_prompt(self, prompt, default):
        if default:
            return f'{prompt} [{default}]: '
        else:
            return f'{prompt}: '


    def edit_items(self, itemclass, *items):
        commands = [['sensible-editor'], ['open', '-W']]
        with NamedTemporaryFile(mode="w+") as tempfile:
            write_items(tempfile, *items)
            tempfile.seek(0)
            command = [os.environ.get('EDITOR')]
            if not command[0] or not shutil.which(command[0]):
                iterator = (c for c in commands if shutil.which(c[0]))
                command = next(filter(None, iterator), None)
                if not command:
                    raise RuntimeError("A text editor at the $EDITOR environment variable is required")
            subprocess.run(command + [tempfile.name])
            tempfile.seek(0)
            new_items = read_items(tempfile, itemclass)
            return new_items



from unittest import TestCase
from unittest.mock import patch

from busy.util.python_version import confirm_python_version

class TestPythonVersion(TestCase):

    @patch('busy.util.python_version.version_info', (3, 0, 0, 'final', 0))
    def test_python_version_fail(self):
        self.assertRaises(RuntimeError, confirm_python_version, (3, 6, 5))

    @patch('busy.util.python_version.version_info', (3, 7, 0, 'final', 0))
    def test_python_version_pass(self):
        confirm_python_version, (3, 6, 5)



import tkinter as tk  # pragma: no cover


class EditTaskWidget(tk.Frame):  # pragma: no cover

    def __init__(self, callback=None, initial_value=''):
        self.callback = callback
        self.root = tk.Tk()
        self.root.state('zoomed')
        main = tk.Frame(self.root)
        main.pack(fill="both", expand=True)
        frame = tk.LabelFrame(main, text="Task")
        self.entry = tk.Entry(frame, width=100)
        self.entry.insert(0, initial_value)
        self.entry.icursor(0)
        self.entry.bind('<Return>', self.enter)
        self.entry.pack()
        self.entry.focus()
        frame.pack(padx=20, pady=20)
        self.root.mainloop()

    def enter(self, event):
        self.value = self.entry.get()
        self.root.destroy()
        if self.callback:
            self.callback(self.value)



from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from unittest import mock
from io import StringIO
from datetime import date as Date

from busy.command.tags_command import TagsCommand
from busy.queue.queue import Queue


# Tired of testing pure units, I present a 2-level integration test

class TestCommandTags(TestCase):

    def test_tags(self):
        q = Queue(items=['a #b', 'c #d', 'e #b'])
        c = TagsCommand()
        c._queue_ = q
        tx = c.call_method()
        self.assertEqual(tx, 'b\nd')



from unittest import TestCase
from unittest.mock import Mock
from unittest.mock import patch
from pathlib import Path

from busy.ui.ui import TerminalUI


class MockSubprocess:

    @staticmethod
    def run(arg):
        Path(arg[-1]).write_text('v')


class MockShUtil:

    @staticmethod
    def which(arg):
        return False


# Not really unit tests, relying on the UI class importing and using other
# stuff.

class TestEditor(TestCase):

    def test_edit_items(self):
        with patch('busy.ui.ui.subprocess', MockSubprocess):
            u = TerminalUI(None)
            ic = Mock()
            ic.schema = ['d']
            i = Mock()
            i.schema = ['d']
            i.d = 'a'
            new = u.edit_items(ic, i)
            ic.create.assert_called_with({'d':'v'})


# A terminal-based UI uses Curses, which offers one-key access to most common
# operations.
#
# The terminal has 3 regions, implemented as curses window objects. The regions
# are organized top-to-bottom with a 1-row gap between them. There is also a
# 1-row/1-column gap around the outside of the main terminal window containing
# the regions.
#
# Status region (_statuswin) - 6 lines at the top. Current queue name, current
# task, some blank space, status of the most recent operation, and input prompt
# for the main command.
#
# List region (_listwin) - Takes up the space between the other two regions.
# Shows mostly lists of tasks for confirmation and/or management.
#
# Edit region (_editwin) - 2 lines at the bottom. Used for editing of individual
# task descriptions, such as when adding a new task to the queue.


import curses
from argparse import Namespace
from string import capwords

from ..ui.ui import TerminalUI

class CursesCancelError(Exception):
    pass

class CursesUI(TerminalUI):  # pragma: nocover

    name = "curses"

    def start(self):
        self._mode = "WORK"
        keynames = self.handler.get_command_values('key', 'name')
        keynames.add(('q', 'quit'))
        skeynames = sorted(keynames, key=lambda kn: kn[0])
        self._main_prompt = skeynames
        curses.wrapper(self.term_loop)

    def output(self, intro=""):
        self._descwin.clear()
        self._descwin.addstr(intro)
        self._descwin.refresh()


    # TODO: Use a better editing component

    def get_string(self, prompt, default=""):
        self._update_status(self.full_prompt(prompt, default))
        curses.echo()
        try:
            string = self._statuswin.getstr()
        except KeyboardInterrupt:
            raise CursesCancelError
        value = string.decode() or default
        curses.noecho()
        return value

    # In Curses, confirm with a single keystroke
    #
    # TODO: Handle the case where the description is too long for the panel

    def get_confirmation(self, verb):
        prompt = f"{verb}? (y/n) --> "
        self._update_status(prompt)
        key = self._get_key()
        confirmed = (key == 'y')
        return confirmed
    
    def term_loop(self, fullwin):
        root = self.handler.root
        self._queue_name = "tasks"
        curses.init_color(curses.COLOR_BLACK, 200, 200, 200)
        curses.init_pair(1, curses.COLOR_CYAN, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        curses.init_pair(3, curses.COLOR_GREEN, curses.COLOR_BLACK)
        h, w = fullwin.getmaxyx()
        self._statuswin = curses.newwin(5, w-2, 1, 1)
        self._descwin = curses.newwin(h-8, w-2, 7, 1)
        self._editwin = curses.newwin(1, w-2, h-2, 1)
        self._status = "Welcome to Busy!"
        list = ""
        cursor = 0
        while True:
            self.output()
            self._update_status(self._main_prompt)
            try:
                key = self._get_key()
            except CommandCanceledError:
                break
            if key == "q":
                break
            command = self.handler.get_command('key', key, self)
            if not command:
                self._status = f"Invalid command {key}"
                continue
            command_name = capwords(command.name)
            self._status = f"{command_name} in progress"
            self._update_status("")
            namespace = Namespace()
            try:
                result = command.execute(namespace)
            except CursesCancelError:
                self._status = f"{command_name} command canceled"
                continue
            self._status = command.status or ""

    def _update_status(self, prompt):
        queue = self.handler.root.get_queue(self._queue_name)
        self._statuswin.clear()
        self._statuswin.move(0, 0)
        self._statuswin.addstr("Queue: ", curses.color_pair(1))
        self._statuswin.addstr(capwords(self._queue_name), curses.A_BOLD)
        self._statuswin.move(1, 0)
        self._statuswin.addstr("Top:   ", curses.color_pair(1))
        self._statuswin.addstr(str(queue.top()) or '', curses.A_BOLD)
        self._statuswin.move(3, 0)
        self._statuswin.addstr(self._status, curses.color_pair(2))
        self._statuswin.move(4, 0)
        if type(prompt) == str:
            self._statuswin.addstr(prompt, curses.color_pair(3))
        elif type(prompt) == list:
            for key, name in prompt:
                pre, it, post = name.partition(key)
                self._statuswin.addstr(pre, curses.color_pair(3))
                self._statuswin.addstr(it, curses.A_UNDERLINE + curses.color_pair(3))
                self._statuswin.addstr(post + " ", curses.color_pair(3))
            self._statuswin.addstr("--> ", curses.color_pair(3))
        cursor = self._statuswin.getyx()
        self._statuswin.refresh()
        self._statuswin.move(*cursor)

    # Convenience method to get one keystroke from the user.
    #     
    def _get_key(self):
        try:
            key = self._statuswin.getkey()
        except KeyboardInterrupt:
            raise CursesCancelError
        self._statuswin.addstr(key)
        return key


    # def listmode(self):
    #     length = len(list.splitlines())
    #     if length:
    #         listpad = curses.newpad(length, curses.COLS)
    #         listpad.addstr(list)
    #         self.listmode = True
    #         listpad.move(cursor, 0)
    #         listpad.addstr(">")
    #         listpad.refresh(0, 0, 5, 0, curses.LINES - 5, curses.COLS)
    #     if self.listmode:
    #         if key == "KEY_UP" and cursor > 0:
    #             cursor = cursor - 1
    #         if key == "KEY_DOWN" and cursor < length:
    #             cursor = cursor + 1



from pathlib import Path
from tempfile import TemporaryDirectory
import os
import subprocess
from pathlib import Path

from busy.model.item import Item
from busy.queue.queue import Queue
from busy.model.item import read_items
from busy.model.item import write_items

class File:

    def __init__(self, path):
        self._path = path

    def read(self, itemclass=Item):
        if self._path.is_file():
            with open(self._path) as datafile:
                return read_items(datafile, itemclass)
        return []

    def save(self, *items):
        if items:
            with open(self._path, 'w') as datafile:
                write_items(datafile, *items)
        else:
            Path(self._path).write_text('')


class FileSystemRoot:

    def __init__(self, path=None):
        if path:
            self._path = Path(path) if isinstance(path, str) else path
            assert isinstance(self._path, Path) and self._path.is_dir()
        else:
            env_var = os.environ.get('BUSY_ROOT')
            self._path = Path(env_var if env_var else Path.home() / '.busy')
            if not self._path.is_dir():
                self._path.mkdir()
        is_git = self.git('rev-parse', check=False)
        if is_git.returncode > 0:
            self.git('init')
        self._files = {}
        self._queues = {}

    @property
    def _str_path(self):
        return str(self._path.resolve())

    def git(self, command, **kwargs):
        return subprocess.run(f'git -C {self._str_path} {command}'.split(), capture_output=True, **kwargs)

    def get_queue(self, name):
        if name not in self._queues:
            queueclass = Queue.family_member('name', name) or Queue
            queuefile = File(self._path / f'{name}.txt')
            self._files[name] = queuefile
            items = queuefile.read(queueclass.itemclass)
            self._queues[name] = queueclass(self, items)
        return self._queues[name]

    def save(self):
        changed = False
        while self._queues:
            key, queue = self._queues.popitem()
            if queue.changed:
                items = queue.all()
                self._files[key].save(*items)
                changed = True
        # if changed:
        #     self.git('add -A')
        #     self.git('commit -m change')

    @property
    def queue_names(self):
        filenames = list(self._path.glob('*.txt'))
        keys = [Path(f).stem for f in filenames]
        return keys



from curses import echo, noecho, newwin
from ..command.command import QueueCommand


class AddCommand(QueueCommand):

    name = 'add'
    key = 'a'

    @classmethod
    def set_parser(self, parser):
        super().set_parser(parser)
        parser.add_argument('--description', nargs='?')

    def clean_args(self):
        super().clean_args()
        if self.is_omitted('description'):
            description = self._ui.get_string("Description")
            self._namespace.description = description
    

    # TODO: Support multiple adds at once    

    def call_method(self):
        description = self._namespace.description
        if description:
            self._queue.add(description)
            self.status = "Added: " + description
        else:
            self.status = "Nothing added"



from .queue import Queue
from ..model.task import Task
from ..model.task import Plan
from ..model.task import DoneTask


# TODO: Move type-specific methods to specific commands

class TodoQueue(Queue):

    itemclass = Task
    name = 'tasks'

    def __init__(self, root=None, items=[]):
        super().__init__(root, items)
        self._plans = None
        self._done = None

    @property
    def plans(self):
        if not self._plans:
            if self.root:
                self._plans = self.root.get_queue(PlanQueue.name)
            else:
                self._plans = PlanQueue()
        return self._plans

    @property
    def done(self):
        if not self._done:
            if self.root:
                self._done = self.root.get_queue(DoneQueue.name)
            else:
                self._done = DoneQueue()
        return self._done


    def defer(self, *indices, date):
        plans = [self.top(i).as_plan(date) for i in indices]
        self.plans.add(*plans)
        self.delete(*indices)
        return plans


    # The indices passed to the activate method are on the plans queue, not this
    # queue.

    def activate(self, *indices):
        tasks = [self.plans.top(i).as_todo() for i in indices]
        added = self.add(*tasks, index=0)
        self.plans.delete(*indices)
        return added


    def finish(self, *indices, date):
        donelist, keeplist = self._split_by_indices(*indices)
        self._items = keeplist
        finished = [t.as_done(date) for t in donelist]
        self.done.add(*finished)
        added = [t for t in [tt.as_followon() for tt in donelist] if t]
        self.add(*added)
        deferred = [t for t in [tt.as_repeat() for tt in donelist] if t]
        self.plans.add(*deferred)
        return (finished, added, deferred)

    def resource(self, index=1):
        return self._items[index-1].resource if self._items else None

    def base(self, index=1):
        return self._items[index-1].base if self._items else None


class PlanQueue(Queue):
    itemclass = Plan
    name = 'plans'


class DoneQueue(Queue):
    itemclass = DoneTask
    name = 'done'



from unittest import TestCase
from unittest.mock import Mock
import unittest
import random

from busy.queue.queue import Queue


class TestQueue(TestCase):

    def test_list(self):
        q = Queue()
        q.add('a')
        q.add('b')
        i = q.list()
        self.assertEqual(i[0][0], 1)
        self.assertEqual(str(i[0][1]), 'a')

    def test_list_with_criteria(self):
        q = Queue()
        q.add('a','b','c')
        i = q.list(1,3)
        self.assertEqual(i[1][0], 3)
        self.assertEqual(str(i[1][1]), 'c')

    def test_select_multiple(self):
        q = Queue()
        q.add('a','b','c')
        t = q.indices(1,3)
        self.assertEqual(len(t), 2)

    def test_pop_multiple(self):
        q = Queue()
        q.add('a','b','c','d')
        q.pop(2,4)
        self.assertEqual(str(q.top(2)), 'd')

    def test_pop_if_nothing(self):
        q = Queue()
        q.pop()
        self.assertEqual(q.count(), 0)

    def test_drop(self):
        q = Queue()
        q.add('a','b','c','d')
        q.drop()
        self.assertEqual(str(q.top(1)), 'b')

    def test_delete(self):
        q = Queue()
        q.add('a','b','c')
        q.delete(2,3)
        self.assertEqual(q.count(), 1)
        self.assertEqual(q._items[0].description, 'a')

    def test_delete_return(self):
        q = Queue()
        q.add('a','b','c')
        x = q.delete(2,3)
        self.assertEqual(len(x), 2)
        self.assertEqual(x[0].description, 'b')

    def test_create_with_array_of_dict(self):
        q = Queue()
        q.add({'description':'a'})
        self.assertEqual(str(q.top()), 'a')

    def test_list_with_range(self):
        q = Queue()
        q.add('a','b','c')
        t = q.indices('2-3')
        self.assertEqual(len(t), 2)

    def test_list_with_range_ending(self):
        q = Queue()
        q.add('a','b','c','d')
        t = q.list('2-')
        self.assertEqual(len(t), 3)
        self.assertEqual(str(t[0][1]), 'b')

    def test_list_only_end(self):
        q = Queue()
        q.add('a', 'b', 'c', 'd')
        t = q.list('-')
        self.assertEqual(len(t), 1)
        self.assertEqual(str(t[0][1]), 'd')

    def test_partial_tags_dont_match(self):
        q = Queue()
        q.add('a #b', 'b #busy')
        t = q.list('b')
        self.assertEqual(len(t), 1)

    def test_tags(self):
        q = Queue()
        q.add('a #b', 'c #d', 'e #b')
        t = q.tags()
        self.assertEqual(t, {'b', 'd'})

    def test_shuffle(self):
        q = Queue()
        q.add('a', 'b', 'c')
        random.seed(1)
        q.shuffle()
        self.assertEqual(q.strings(1,2,3), ['c','a','b'])

    def test_replace(self):
        q = Queue()
        q.add('a', 'b', 'c')
        i = Mock()
        i.__str__ = lambda x:'x'
        q.replace([2], [i])
        s = q.list()
        self.assertEqual(str(s[1][1]), 'x')

